## Copyright (C) 2024  Solaar Contributors https://pwr-solaar.github.io/Solaar/
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""The tests work by creating a faked device (from the hidpp module) that uses provided data as responses to HID++ commands.
The device uses some methods from the real device to set up data structures that are needed for some tests.
"""

from dataclasses import dataclass
from typing import Any

import pytest

from logitech_receiver import common
from logitech_receiver import hidpp20
from logitech_receiver import hidpp20_constants
from logitech_receiver import settings_templates
from logitech_receiver import special_keys

from . import fake_hidpp

# TODO action part of DpiSlidingXY, MouseGesturesXY


class Setup:
    def __init__(self, test, *params):
        self.test = test
        self.responses = [r for r in params if isinstance(r, fake_hidpp.Response)]
        self.choices = None if isinstance(params[0], fake_hidpp.Response) else params[0]


@dataclass
class RegisterTest:
    sclass: Any
    initial_value: Any = False
    write_value: Any = True
    write_params: str = "01"


register_tests = [
    Setup(
        RegisterTest(settings_templates.RegisterHandDetection, False, True, [b"\x00\x00\x00"]),
        fake_hidpp.Response("000030", 0x8101),  # keyboard_hand_detection
        fake_hidpp.Response("000000", 0x8001, "000000"),
    ),
    Setup(
        RegisterTest(settings_templates.RegisterHandDetection, True, False, [b"\x00\x00\x30"]),
        fake_hidpp.Response("000000", 0x8101),  # keyboard_hand_detection
        fake_hidpp.Response("000030", 0x8001, "000030"),
    ),
    Setup(
        RegisterTest(settings_templates.RegisterSmoothScroll, False, True, [b"\x40"]),
        fake_hidpp.Response("00", 0x8101),  # mouse_button_flags
        fake_hidpp.Response("40", 0x8001, "40"),
    ),
    Setup(
        RegisterTest(settings_templates.RegisterSideScroll, True, False, [b"\x00"]),
        fake_hidpp.Response("02", 0x8101),  # mouse_button_flags
        fake_hidpp.Response("00", 0x8001, "00"),
    ),
    Setup(
        RegisterTest(settings_templates.RegisterFnSwap, False, True, [b"\x00\x01"]),
        fake_hidpp.Response("0000", 0x8109),  # keyboard_fn_swap
        fake_hidpp.Response("0001", 0x8009, "0001"),
    ),
    Setup(
        RegisterTest(
            settings_templates._PerformanceMXDpi, common.NamedInt(0x88, "800"), common.NamedInt(0x89, "900"), [b"\x89"]
        ),
        fake_hidpp.Response("88", 0x8163),  # mouse_dpi
        fake_hidpp.Response("89", 0x8063, "89"),
    ),
]


@pytest.mark.parametrize("test", register_tests)
def test_register_template(test, mocker):
    device = fake_hidpp.Device(protocol=1.0, responses=test.responses)
    spy_request = mocker.spy(device, "request")

    setting = test.test.sclass.build(device)
    value = setting.read(cached=False)
    cached_value = setting.read(cached=True)
    write_value = setting.write(test.test.write_value)

    assert setting is not None
    assert value == test.test.initial_value
    assert cached_value == test.test.initial_value
    assert write_value == test.test.write_value
    spy_request.assert_called_with(test.test.sclass.register + 0x8000, *test.test.write_params)


@dataclass
class FeatureTest:
    sclass: Any
    initial_value: Any = False
    write_value: Any = True
    matched_calls: int = 1
    offset: int = 0x04
    version: int = 0x00
    rewrite: bool = False
    readable: bool = True


simple_tests = [
    Setup(
        FeatureTest(settings_templates.K375sFnSwap, False, True, offset=0x06),
        fake_hidpp.Response("FF0001", 0x0600, "FF"),
        fake_hidpp.Response("FF0101", 0x0610, "FF01"),
    ),
    Setup(
        FeatureTest(settings_templates.K375sFnSwap, False, True, offset=0x06),
        fake_hidpp.Response("050001", 0x0000, "1815"),  # HOSTS_INFO
        fake_hidpp.Response("FF0001", 0x0600, "FF"),
        fake_hidpp.Response("FF0101", 0x0610, "FF01"),
    ),
    Setup(
        FeatureTest(settings_templates.K375sFnSwap, False, True, offset=0x06),
        fake_hidpp.Response("050001", 0x0000, "1815"),  # HOSTS_INFO
        fake_hidpp.Response("07050301", 0x0500),  # current host is 0x01, i.e., host 2
        fake_hidpp.Response("010001", 0x0600, "01"),
        fake_hidpp.Response("010101", 0x0610, "0101"),
    ),
    Setup(
        FeatureTest(settings_templates.FnSwap, True, False),
        fake_hidpp.Response("01", 0x0400),
        fake_hidpp.Response("00", 0x0410, "00"),
    ),
    Setup(
        FeatureTest(settings_templates.NewFnSwap, True, False),
        fake_hidpp.Response("01", 0x0400),
        fake_hidpp.Response("00", 0x0410, "00"),
    ),
    #    Setup(  # Backlight has caused problems
    #        FeatureTest(settings_templates.Backlight, 0, 5, offset=0x06),
    #        fake_hidpp.Response("00", 0x0600),
    #        fake_hidpp.Response("05", 0x0610, "05"),
    #    ),
    Setup(
        FeatureTest(settings_templates.Backlight2DurationHandsOut, 80, 160, version=0x03),
        fake_hidpp.Response("011830000000100040006000", 0x0400),
        fake_hidpp.Response("0118FF00200040006000", 0x0410, "0118FF00200040006000"),
    ),
    Setup(
        FeatureTest(settings_templates.Backlight2DurationHandsIn, 320, 160, version=0x03),
        fake_hidpp.Response("011830000000200040006000", 0x0400),
        fake_hidpp.Response("0118FF00200020006000", 0x0410, "0118FF00200020006000"),
    ),
    Setup(
        FeatureTest(settings_templates.Backlight2DurationPowered, 480, 80, version=0x03),
        fake_hidpp.Response("011830000000200040006000", 0x0400),
        fake_hidpp.Response("0118FF00200040001000", 0x0410, "0118FF00200040001000"),
    ),
    Setup(
        FeatureTest(settings_templates.Backlight3, 0x50, 0x70),
        fake_hidpp.Response("50", 0x0410),
        fake_hidpp.Response("70", 0x0420, "007009"),
    ),
    Setup(
        FeatureTest(settings_templates.HiResScroll, True, False),
        fake_hidpp.Response("01", 0x0400),
        fake_hidpp.Response("00", 0x0410, "00"),
    ),
    Setup(
        FeatureTest(settings_templates.LowresMode, False, True),
        fake_hidpp.Response("00", 0x0400),
        fake_hidpp.Response("01", 0x0410, "01"),
    ),
    Setup(
        FeatureTest(settings_templates.HiresSmoothInvert, True, False),
        fake_hidpp.Response("06", 0x0410),
        fake_hidpp.Response("02", 0x0420, "02"),
    ),
    Setup(
        FeatureTest(settings_templates.HiresSmoothResolution, True, False),
        fake_hidpp.Response("06", 0x0410),
        fake_hidpp.Response("04", 0x0420, "04"),
    ),
    Setup(
        FeatureTest(settings_templates.HiresMode, False, True),
        fake_hidpp.Response("06", 0x0410),
        fake_hidpp.Response("07", 0x0420, "07"),
    ),
    Setup(
        FeatureTest(settings_templates.PointerSpeed, 0x0100, 0x0120),
        fake_hidpp.Response("0100", 0x0400),
        fake_hidpp.Response("0120", 0x0410, "0120"),
    ),
    Setup(
        FeatureTest(settings_templates.ThumbMode, True, False),
        fake_hidpp.Response("0100", 0x0410),
        fake_hidpp.Response("0000", 0x0420, "0000"),
    ),
    Setup(
        FeatureTest(settings_templates.ThumbInvert, False, True),
        fake_hidpp.Response("0100", 0x0410),
        fake_hidpp.Response("0101", 0x0420, "0101"),
    ),
    Setup(
        FeatureTest(settings_templates.DivertCrown, False, True),
        fake_hidpp.Response("01", 0x0410),
        fake_hidpp.Response("02", 0x0420, "02"),
    ),
    Setup(
        FeatureTest(settings_templates.CrownSmooth, True, False),
        fake_hidpp.Response("0001", 0x0410),
        fake_hidpp.Response("0002", 0x0420, "0002"),
    ),
    Setup(
        FeatureTest(settings_templates.DivertGkeys, False, True),
        fake_hidpp.Response("01", 0x0420, "01"),
    ),
    Setup(
        FeatureTest(settings_templates.ScrollRatchet, 2, 1),
        fake_hidpp.Response("02", 0x0400),
        fake_hidpp.Response("01", 0x0410, "01"),
    ),
    Setup(
        FeatureTest(settings_templates.SmartShift, 1, 10),
        fake_hidpp.Response("0100", 0x0400),
        fake_hidpp.Response("000A", 0x0410, "000A"),
    ),
    Setup(
        FeatureTest(settings_templates.SmartShift, 5, 50),
        fake_hidpp.Response("0005", 0x0400),
        fake_hidpp.Response("00FF", 0x0410, "00FF"),
    ),
    Setup(
        FeatureTest(settings_templates.SmartShiftEnhanced, 5, 50),
        fake_hidpp.Response("0005", 0x0410),
        fake_hidpp.Response("00FF", 0x0420, "00FF"),
    ),
    Setup(
        FeatureTest(settings_templates.DisableKeyboardKeys, {1: True, 8: True}, {1: False, 8: True}),
        fake_hidpp.Response("09", 0x0400),
        fake_hidpp.Response("09", 0x0410),
        fake_hidpp.Response("08", 0x0420, "08"),
    ),
    Setup(
        FeatureTest(settings_templates.DualPlatform, 0, 1),
        fake_hidpp.Response("00", 0x0400),
        fake_hidpp.Response("01", 0x0420, "01"),
    ),
    Setup(
        FeatureTest(settings_templates.MKeyLEDs, {1: False, 2: False, 4: False}, {1: False, 2: True, 4: True}),
        fake_hidpp.Response("03", 0x0400),
        fake_hidpp.Response("06", 0x0410, "06"),
    ),
    Setup(
        FeatureTest(settings_templates.MRKeyLED, False, True),
        fake_hidpp.Response("01", 0x0400, "01"),
    ),
    Setup(
        FeatureTest(settings_templates.Sidetone, 5, 0xA),
        fake_hidpp.Response("05", 0x0400),
        fake_hidpp.Response("0A", 0x0410, "0A"),
    ),
    Setup(
        FeatureTest(settings_templates.ADCPower, 5, 0xA, version=0x03),
        fake_hidpp.Response("05", 0x0410),
        fake_hidpp.Response("0A", 0x0420, "0A"),
    ),
    Setup(
        FeatureTest(settings_templates.LEDControl, 0, 1),
        fake_hidpp.Response("00", 0x0470),
        fake_hidpp.Response("01", 0x0480, "01"),
    ),
    Setup(
        FeatureTest(
            settings_templates.LEDZoneSetting,
            hidpp20.LEDEffectSetting(ID=3, intensity=0x50, period=0x100),
            hidpp20.LEDEffectSetting(ID=3, intensity=0x50, period=0x101),
        ),
        fake_hidpp.Response("0100000001", 0x0400),
        fake_hidpp.Response("00000102", 0x0410, "00FF00"),
        fake_hidpp.Response("0000000300040005", 0x0420, "000000"),
        fake_hidpp.Response("0001000B00080009", 0x0420, "000100"),
        fake_hidpp.Response("000000000000010050", 0x04E0, "00"),
        fake_hidpp.Response("000000000000000101500000", 0x0430, "000000000000000101500000"),
    ),
    Setup(
        FeatureTest(settings_templates.RGBControl, 0, 1),
        fake_hidpp.Response("0000", 0x0450),
        fake_hidpp.Response("010100", 0x0450, "0101"),
    ),
    Setup(
        FeatureTest(
            settings_templates.RGBEffectSetting,
            hidpp20.LEDEffectSetting(ID=3, intensity=0x50, period=0x100),
            hidpp20.LEDEffectSetting(ID=2, color=0x505050, speed=0x50),
            readable=False,
        ),
        fake_hidpp.Response("FFFF0100000001", 0x0400, "FFFF00"),
        fake_hidpp.Response("0000000102", 0x0400, "00FF00"),
        fake_hidpp.Response("0000000300040005", 0x0400, "000000"),
        fake_hidpp.Response("0001000200080009", 0x0400, "000100"),
        fake_hidpp.Response("000000000000010050", 0x04E0, "00"),
        fake_hidpp.Response("00015050505000000000000001", 0x0410, "00015050505000000000000001"),
    ),
    Setup(
        FeatureTest(
            settings_templates.RGBEffectSetting,
            None,
            hidpp20.LEDEffectSetting(ID=3, intensity=0x60, period=0x101),
            readable=False,
        ),
        fake_hidpp.Response("FFFF0100000001", 0x0400, "FFFF00"),
        fake_hidpp.Response("0000000102", 0x0400, "00FF00"),
        fake_hidpp.Response("0000000300040005", 0x0400, "000000"),
        fake_hidpp.Response("0001000200080009", 0x0400, "000100"),
        fake_hidpp.Response("00000000000000010160000001", 0x0410, "00000000000000010160000001"),
    ),
    Setup(
        FeatureTest(
            settings_templates.RGBEffectSetting,
            None,
            hidpp20.LEDEffectSetting(ID=3, intensity=0x60, period=0x101),
            readable=False,
        ),
        fake_hidpp.Response("FF000200020004000000000000000000", 0x0400, "FFFF00"),
        fake_hidpp.Response("00000002040000000000000000000000", 0x0400, "00FF00"),
        fake_hidpp.Response("00000000000000000000000000000000", 0x0400, "000000"),
        fake_hidpp.Response("00010001000000000000000000000000", 0x0400, "000100"),
        fake_hidpp.Response("00020003C00503E00000000000000000", 0x0400, "000200"),
        fake_hidpp.Response("0003000AC0011E0B0000000000000000", 0x0400, "000300"),
        fake_hidpp.Response("01000001070000000000000000000000", 0x0400, "01FF00"),
        fake_hidpp.Response("01000000000000000000000000000000", 0x0400, "010000"),
        fake_hidpp.Response("01010001000000000000000000000000", 0x0400, "010100"),
        fake_hidpp.Response("0102000AC0011E0B0000000000000000", 0x0400, "010200"),
        fake_hidpp.Response("01030003C00503E00000000000000000", 0x0400, "010300"),
        fake_hidpp.Response("01040004DCE1001E0000000000000000", 0x0400, "010400"),
        fake_hidpp.Response("0105000B000000320000000000000000", 0x0400, "010500"),
        fake_hidpp.Response("0106000C001B02340000000000000000", 0x0400, "010600"),
        fake_hidpp.Response("00020000000000010160000001", 0x0410, "00020000000000010160000001"),
    ),
    Setup(
        FeatureTest(settings_templates.Backlight2, 0xFF, 0x00),
        common.NamedInts(Disabled=0xFF, Enabled=0x00),
        fake_hidpp.Response("000201000000000000000000", 0x0400),
        fake_hidpp.Response("010201", 0x0410, "0102FF00000000000000"),
    ),
    Setup(
        FeatureTest(settings_templates.Backlight2, 0x03, 0xFF),
        common.NamedInts(Disabled=0xFF, Automatic=0x01, Manual=0x03),
        fake_hidpp.Response("011838000000000000000000", 0x0400),
        fake_hidpp.Response("001801", 0x0410, "0018FF00000000000000"),
    ),
    Setup(
        FeatureTest(settings_templates.Backlight2Level, 0, 3, version=0x03),
        [0, 4],
        fake_hidpp.Response("011830000000000000000000", 0x0400),
        fake_hidpp.Response("05", 0x0420),
        fake_hidpp.Response("01180103000000000000", 0x0410, "0118FF03000000000000"),
    ),
    Setup(
        FeatureTest(settings_templates.Backlight2Level, 0, 2, version=0x03),
        [0, 4],
        fake_hidpp.Response("011830000000000000000000", 0x0400),
        fake_hidpp.Response("05", 0x0420),
        fake_hidpp.Response("01180102000000000000", 0x0410, "0118FF02000000000000"),
    ),
    Setup(
        FeatureTest(settings_templates.OnboardProfiles, 0, 1, offset=0x0C),
        common.NamedInts(**{"Disabled": 0, "Profile 1": 1, "Profile 2": 2}),
        fake_hidpp.Response("01030001010101000101", 0x0C00),
        fake_hidpp.Response("00010100000201FFFFFFFFFFFFFFFFFF", 0x0C50, "00000000"),
        fake_hidpp.Response("000201FFFFFFFFFFFFFFFFFFFFFFFFFF", 0x0C50, "00000004"),
        fake_hidpp.Response("FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF", 0x0C50, "00000008"),
        fake_hidpp.Response("02", 0x0C20),
        fake_hidpp.Response("01", 0x0C10, "01"),
        fake_hidpp.Response("0001", 0x0C30, "0001"),
    ),
    Setup(
        FeatureTest(settings_templates.OnboardProfiles, 1, 0, offset=0x0C),
        common.NamedInts(**{"Disabled": 0, "Profile 1": 1, "Profile 2": 2}),
        fake_hidpp.Response("01030001010101000101", 0x0C00),
        fake_hidpp.Response("00010100000201FFFFFFFFFFFFFFFFFF", 0x0C50, "00000000"),
        fake_hidpp.Response("000201FFFFFFFFFFFFFFFFFFFFFFFFFF", 0x0C50, "00000004"),
        fake_hidpp.Response("FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF", 0x0C50, "00000008"),
        fake_hidpp.Response("01", 0x0C20),
        fake_hidpp.Response("0001", 0x0C40),
        fake_hidpp.Response("02", 0x0C10, "02"),
    ),
    Setup(
        FeatureTest(settings_templates.ReportRate, 1, 5, offset=0x0C),
        common.NamedInts(**{"1ms": 1, "2ms": 2, "5ms": 5, "6ms": 6}),
        fake_hidpp.Response("33", 0x0C00),
        fake_hidpp.Response("01", 0x0C10),
        fake_hidpp.Response("05", 0x0C20, "05"),
    ),
    Setup(
        FeatureTest(settings_templates.ExtendedReportRate, 1, 5, offset=0x0C),
        common.NamedInts(**{"8ms": 0, "4ms": 1, "500us": 4, "250us": 5}),
        fake_hidpp.Response("33", 0x0C10),
        fake_hidpp.Response("01", 0x0C20),
        fake_hidpp.Response("05", 0x0C30, "05"),
    ),
    Setup(
        FeatureTest(settings_templates.AdjustableDpi, 800, 400, version=0x03),
        common.NamedInts.list([400, 800, 1600]),
        fake_hidpp.Response("000190032006400000", 0x0410, "000000"),
        fake_hidpp.Response("000320", 0x0420),
        fake_hidpp.Response("000190", 0x0430, "000190"),
    ),
    Setup(
        FeatureTest(settings_templates.AdjustableDpi, 256, 512, version=0x03),
        common.NamedInts.list([256, 512]),
        fake_hidpp.Response("000100e10002000000", 0x0410, "000000"),
        fake_hidpp.Response("000100", 0x0420),
        fake_hidpp.Response("000200", 0x0430, "000200"),
    ),
    Setup(
        FeatureTest(settings_templates.AdjustableDpi, 400, 800, version=0x03),
        common.NamedInts.list([400, 800, 1200, 1600]),
        fake_hidpp.Response("000190E19006400000000000000000", 0x0410, "000000"),
        fake_hidpp.Response("000190", 0x0420),
        fake_hidpp.Response("000320", 0x0430, "000320"),
    ),
    Setup(
        FeatureTest(settings_templates.Multiplatform, 0, 1),
        common.NamedInts(**{"MacOS 0.1-0.5": 0, "iOS 0.1-0.7": 1, "Linux 0.2-0.9": 2, "Windows 0.3-0.9": 3}),
        fake_hidpp.Response("020004000001", 0x0400),
        fake_hidpp.Response("00FF200000010005", 0x0410, "00"),
        fake_hidpp.Response("01FF400000010007", 0x0410, "01"),
        fake_hidpp.Response("02FF040000020009", 0x0410, "02"),
        fake_hidpp.Response("03FF010000030009", 0x0410, "03"),
        fake_hidpp.Response("FF01", 0x0430, "FF01"),
    ),
    Setup(
        FeatureTest(settings_templates.ChangeHost, 1, 0),
        common.NamedInts(**{"1:ABCDEF": 0, "2:GHIJKL": 1}),
        fake_hidpp.Response("050003", 0x0000, "1815"),  # HOSTS_INFO
        fake_hidpp.Response("01000200", 0x0500),
        fake_hidpp.Response("000100000600", 0x0510, "00"),
        fake_hidpp.Response("000041424344454600", 0x0530, "0000"),
        fake_hidpp.Response("000100000600", 0x0510, "01"),
        fake_hidpp.Response("00004748494A4B4C00", 0x0530, "0100"),
        fake_hidpp.Response("0201", 0x0400),
        fake_hidpp.Response(True, 0x0410, "00"),
    ),
    Setup(
        FeatureTest(settings_templates.BrightnessControl, 0x10, 0x20),
        [0, 80],
        fake_hidpp.Response("00505100000000", 0x0400),  # 0 to 80, all acceptable, no separate on/off
        fake_hidpp.Response("10", 0x0410),  # brightness 16
        fake_hidpp.Response("0020", 0x0420, "0020"),  # set brightness 32
    ),
    Setup(
        FeatureTest(settings_templates.BrightnessControl, 0x10, 0x00),
        [0, 80],
        fake_hidpp.Response("00505104000000", 0x0400),  # 0 to 80, all acceptable, separate on/off
        fake_hidpp.Response("10", 0x0410),  # brightness 16
        fake_hidpp.Response("01", 0x0430),  # on
        fake_hidpp.Response("00", 0x0440),  # set off
        fake_hidpp.Response("0000", 0x0420, "0000"),  # set brightness 0
    ),
    Setup(
        FeatureTest(settings_templates.BrightnessControl, 0x00, 0x20),
        [0, 80],
        fake_hidpp.Response("00505104000000", 0x0400),  # 0 to 80, all acceptable, separate on/off
        fake_hidpp.Response("10", 0x0410),  # brightness 16
        fake_hidpp.Response("00", 0x0430),  # off
        fake_hidpp.Response("01", 0x0440),  # set on
        fake_hidpp.Response("0020", 0x0420, "0020"),  # set brightness 32
    ),
    Setup(
        FeatureTest(settings_templates.BrightnessControl, 0x20, 0x08),
        [0, 80],
        fake_hidpp.Response("00504104001000", 0x0400),  # 16 to 80, all acceptable, separate on/off
        fake_hidpp.Response("20", 0x0410),  # brightness 32
        fake_hidpp.Response("01", 0x0430),  # on
        fake_hidpp.Response("00", 0x0440, "00"),  # set off
    ),
    Setup(
        FeatureTest(settings_templates.SpeedChange, 0, None, 0),  # need to set up all settings to successfully write
        common.NamedInts(**{"Off": 0, "DPI Change": 0xED}),
        *fake_hidpp.responses_speedchange,
    ),
]


@pytest.fixture
def mock_gethostname(mocker):
    mocker.patch("socket.gethostname", return_value="ABCDEF.foo.org")


@pytest.mark.parametrize("test", simple_tests)
def test_simple_template(test, mocker, mock_gethostname):
    tst = test.test
    print("TEST", tst.sclass, tst.sclass.feature)
    device = fake_hidpp.Device(responses=test.responses, feature=tst.sclass.feature, offset=tst.offset, version=tst.version)
    spy_request = mocker.spy(device, "request")

    setting = settings_templates.check_feature(device, tst.sclass)
    assert setting is not None
    if isinstance(setting, list):
        setting = setting[0]
    if isinstance(test.choices, list):
        assert setting._validator.min_value == test.choices[0]
        assert setting._validator.max_value == test.choices[1]
    elif test.choices is not None:
        assert setting.choices == test.choices

    if tst.readable:
        value = setting.read(cached=False)
        assert value == tst.initial_value

        cached_value = setting.read(cached=True)
        assert cached_value == tst.initial_value

    write_value = setting.write(tst.write_value) if tst.write_value is not None else None
    assert write_value == tst.write_value

    fake_hidpp.match_requests(tst.matched_calls, test.responses, spy_request.call_args_list)


responses_reprog_controls = [
    fake_hidpp.Response("03", 0x0500),
    fake_hidpp.Response("00500038010001010400000000000000", 0x0510, "00"),  # left button
    fake_hidpp.Response("00510039010001010400000000000000", 0x0510, "01"),  # right button
    fake_hidpp.Response("00C4009D310003070500000000000000", 0x0510, "02"),  # smart shift
    fake_hidpp.Response("00500000000000000000000000000000", 0x0520, "0050"),  # left button current
    fake_hidpp.Response("00510000500000000000000000000000", 0x0520, "0051"),  # right button current
    fake_hidpp.Response("00C40000000000000000000000000000", 0x0520, "00C4"),  # smart shift current
    fake_hidpp.Response("00500005000000000000000000000000", 0x0530, "0050000050"),  # left button write
    fake_hidpp.Response("00510005000000000000000000000000", 0x0530, "0051000050"),  # right button write
    fake_hidpp.Response("00C4000C400000000000000000000000", 0x0530, "00C40000C4"),  # smart shift write
]

key_tests = [
    Setup(
        FeatureTest(settings_templates.ReprogrammableKeys, {0x50: 0x50, 0x51: 0x50, 0xC4: 0xC4}, {0x51: 0x51}, 4, offset=0x05),
        {
            common.NamedInt(0x50, "Left Button"): common.UnsortedNamedInts(Left_Click=0x50, Right_Click=0x51),
            common.NamedInt(0x51, "Right Button"): common.UnsortedNamedInts(Right_Click=0x51, Left_Click=0x50),
            common.NamedInt(0xC4, "Smart Shift"): common.UnsortedNamedInts(Smart_Shift=0xC4, Left_Click=80, Right_Click=81),
        },
        *responses_reprog_controls,
        fake_hidpp.Response("0051000051", 0x0530, "0051000051"),  # right button set write
    ),
    Setup(
        FeatureTest(settings_templates.DivertKeys, {0xC4: 0}, {0xC4: 1}, 2, offset=0x05),
        {common.NamedInt(0xC4, "Smart Shift"): common.NamedInts(Regular=0, Diverted=1, Mouse_Gestures=2)},
        *responses_reprog_controls,
        fake_hidpp.Response("00C4020000", 0x0530, "00C4020000"),  # Smart Shift write
        fake_hidpp.Response("00C4030000", 0x0530, "00C4030000"),  # Smart Shift divert write
    ),
    Setup(
        FeatureTest(settings_templates.DivertKeys, {0xC4: 0}, {0xC4: 2}, 2, offset=0x05),
        {common.NamedInt(0xC4, "Smart Shift"): common.NamedInts(Regular=0, Diverted=1, Mouse_Gestures=2, Sliding_DPI=3)},
        *responses_reprog_controls,
        fake_hidpp.Response("0A0001", 0x0000, "2201"),  # ADJUSTABLE_DPI
        fake_hidpp.Response("00C4300000", 0x0530, "00C4300000"),  # Smart Shift write
        fake_hidpp.Response("00C4030000", 0x0530, "00C4030000"),  # Smart Shift divert write
    ),
    Setup(
        FeatureTest(settings_templates.PersistentRemappableAction, {80: 16797696, 81: 16797696}, {0x51: 16797952}, 3),
        {
            common.NamedInt(80, "Left Button"): special_keys.KEYS_KEYS_CONSUMER,
            common.NamedInt(81, "Right Button"): special_keys.KEYS_KEYS_CONSUMER,
        },
        fake_hidpp.Response("050001", 0x0000, "1B04"),  # REPROG_CONTROLS_V4
        *responses_reprog_controls,
        fake_hidpp.Response("0041", 0x0400),
        fake_hidpp.Response("0201", 0x0410),
        fake_hidpp.Response("02", 0x0400),
        fake_hidpp.Response("0050", 0x0420, "00FF"),  # left button
        fake_hidpp.Response("0051", 0x0420, "01FF"),  # right button
        fake_hidpp.Response("0050000100500000", 0x0430, "0050FF"),  # left button current
        fake_hidpp.Response("0051000100500001", 0x0430, "0051FF"),  # right button current
        fake_hidpp.Response("0050FF01005000", 0x0440, "0050FF01005000"),  # left button write
        fake_hidpp.Response("0051FF01005000", 0x0440, "0051FF01005000"),  # right button write
        fake_hidpp.Response("0051FF01005100", 0x0440, "0051FF01005100"),  # right button set write
    ),
    Setup(
        FeatureTest(
            settings_templates.Gesture2Gestures,
            {
                1: True,
                2: True,
                30: True,
                10: True,
                45: False,
                42: True,
                43: True,
                64: False,
                65: False,
                67: False,
                84: True,
                34: False,
            },
            {45: True},
            4,
        ),
        *fake_hidpp.responses_gestures,
        fake_hidpp.Response("0001FF6F", 0x0420, "0001FF6F"),  # write
        fake_hidpp.Response("01010F04", 0x0420, "01010F04"),
        fake_hidpp.Response("0001FF7F", 0x0420, "0001FF7F"),  # write 45
        fake_hidpp.Response("01010F04", 0x0420, "01010F04"),
    ),
    Setup(
        FeatureTest(
            settings_templates.Gesture2Divert,
            {1: False, 2: False, 10: False, 44: False, 64: False, 65: False, 67: False, 84: False, 85: False, 100: False},
            {44: True},
            4,
        ),
        *fake_hidpp.responses_gestures,
        fake_hidpp.Response("0001FF00", 0x0440, "0001FF00"),  # write
        fake_hidpp.Response("01010300", 0x0440, "01010300"),
        fake_hidpp.Response("0001FF08", 0x0440, "0001FF08"),  # write 44
        fake_hidpp.Response("01010300", 0x0440, "01010300"),
    ),
    Setup(
        FeatureTest(settings_templates.Gesture2Params, {4: {"scale": 256}}, {4: {"scale": 128}}, 2),
        *fake_hidpp.responses_gestures,
        fake_hidpp.Response("000100FF000000000000000000000000", 0x0480, "000100FF"),
        fake_hidpp.Response("000080FF000000000000000000000000", 0x0480, "000080FF"),
    ),
    Setup(
        FeatureTest(settings_templates.Equalizer, {0: -0x20, 1: 0x10}, {1: 0x18}, 2),
        [-32, 32],
        fake_hidpp.Response("0220000000", 0x0400),
        fake_hidpp.Response("0000800100000000000000", 0x0410, "00"),
        fake_hidpp.Response("E010", 0x0420, "00"),
        fake_hidpp.Response("E010", 0x0430, "02E010"),
        fake_hidpp.Response("E018", 0x0430, "02E018"),
    ),
    Setup(
        FeatureTest(settings_templates.PerKeyLighting, {1: -1, 2: -1, 9: -1, 10: -1, 113: -1}, {2: 0xFF0000}, 4, 4, 0, 1),
        {
            common.NamedInt(1, "A"): special_keys.COLORSPLUS,
            common.NamedInt(2, "B"): special_keys.COLORSPLUS,
            common.NamedInt(9, "I"): special_keys.COLORSPLUS,
            common.NamedInt(10, "J"): special_keys.COLORSPLUS,
            common.NamedInt(113, "KEY 113"): special_keys.COLORSPLUS,
        },
        fake_hidpp.Response("00000606000000000000000000000000", 0x0400, "0000"),  # first group of keys
        fake_hidpp.Response("00000200000000000000000000000000", 0x0400, "0001"),  # second group of keys
        fake_hidpp.Response("00000000000000000000000000000000", 0x0400, "0002"),  # last group of keys
        fake_hidpp.Response("02FF0000", 0x0410, "02FF0000"),  # write one value
        fake_hidpp.Response("00", 0x0470, "00"),  # finish
        fake_hidpp.Response("02FF0000", 0x0410, "02FF0000"),  # write one value
        fake_hidpp.Response("00", 0x0470, "00"),  # finish
    ),
    Setup(
        FeatureTest(
            settings_templates.PerKeyLighting,
            {1: -1, 2: -1, 9: -1, 10: -1, 113: -1},
            {2: 0xFF0000, 9: 0xFF0000, 10: 0xFF0000},
            8,
            4,
            0,
            1,
        ),
        {
            common.NamedInt(1, "A"): special_keys.COLORSPLUS,
            common.NamedInt(2, "B"): special_keys.COLORSPLUS,
            common.NamedInt(9, "I"): special_keys.COLORSPLUS,
            common.NamedInt(10, "J"): special_keys.COLORSPLUS,
            common.NamedInt(113, "KEY 113"): special_keys.COLORSPLUS,
        },
        fake_hidpp.Response("00000606000000000000000000000000", 0x0400, "0000"),  # first group of keys
        fake_hidpp.Response("00000200000000000000000000000000", 0x0400, "0001"),  # second group of keys
        fake_hidpp.Response("00000000000000000000000000000000", 0x0400, "0002"),  # last group of keys
        fake_hidpp.Response("02FF0000", 0x0410, "02FF0000"),  # write one value
        fake_hidpp.Response("00", 0x0470, "00"),  # finish
        fake_hidpp.Response("09FF0000", 0x0410, "09FF0000"),  # write one value
        fake_hidpp.Response("00", 0x0470, "00"),  # finish
        fake_hidpp.Response("0AFF0000", 0x0410, "0AFF0000"),  # write one value
        fake_hidpp.Response("00", 0x0470, "00"),  # finish
        fake_hidpp.Response("02FF000009FF00000AFF0000", 0x0410, "02FF000009FF00000AFF0000"),  # write three values
        fake_hidpp.Response("00", 0x0470, "00"),  # finish
    ),
    Setup(
        FeatureTest(
            settings_templates.PerKeyLighting,
            {1: -1, 2: -1, 9: -1, 10: -1, 113: -1, 114: -1},
            {1: 0xFF0000, 2: 0xFF0000, 9: 0xFF0000, 10: 0xFF0000, 113: 0x00FF00, 114: 0xFF0000},
            15,
            4,
            0,
            1,
        ),
        {
            common.NamedInt(1, "A"): special_keys.COLORSPLUS,
            common.NamedInt(2, "B"): special_keys.COLORSPLUS,
            common.NamedInt(9, "I"): special_keys.COLORSPLUS,
            common.NamedInt(10, "J"): special_keys.COLORSPLUS,
            common.NamedInt(113, "KEY 113"): special_keys.COLORSPLUS,
            common.NamedInt(114, "KEY 114"): special_keys.COLORSPLUS,
        },
        fake_hidpp.Response("00000606000000000000000000000000", 0x0400, "0000"),  # first group of keys
        fake_hidpp.Response("00000600000000000000000000000000", 0x0400, "0001"),  # second group of keys
        fake_hidpp.Response("00000000000000000000000000000000", 0x0400, "0002"),  # last group of keys
        fake_hidpp.Response("01FF0000", 0x0410, "01FF0000"),  # write one value
        fake_hidpp.Response("00", 0x0470, "00"),  # finish
        fake_hidpp.Response("02FF0000", 0x0410, "02FF0000"),  # write one value
        fake_hidpp.Response("00", 0x0470, "00"),  # finish
        fake_hidpp.Response("09FF0000", 0x0410, "09FF0000"),  # write one value
        fake_hidpp.Response("00", 0x0470, "00"),  # finish
        fake_hidpp.Response("0AFF0000", 0x0410, "0AFF0000"),  # write one value
        fake_hidpp.Response("00", 0x0470, "00"),  # finish
        fake_hidpp.Response("7100FF00", 0x0410, "7100FF00"),  # write one value
        fake_hidpp.Response("00", 0x0470, "00"),  # finish
        fake_hidpp.Response("72FF0000", 0x0410, "72FF0000"),  # write one value
        fake_hidpp.Response("00", 0x0470, "00"),  # finish
        fake_hidpp.Response("FF00000102090A72", 0x460, "FF00000102090A72"),  # write one value for five keys
        fake_hidpp.Response("7100FF00", 0x0410, "7100FF00"),  # write one value
        fake_hidpp.Response("00", 0x0470, "00"),  # finish
    ),
    Setup(
        FeatureTest(settings_templates.ExtendedAdjustableDpi, {0: 256}, {0: 512}, 2, offset=0x9),
        {common.NamedInt(0, "X"): common.NamedInts.list([256, 512])},
        fake_hidpp.Response("000000", 0x0910, "00"),  # no y direction, no lod
        fake_hidpp.Response("0000000100e10002000000", 0x0920, "000000"),
        fake_hidpp.Response("00010000000000000000", 0x0950),
        fake_hidpp.Response("000100000000", 0x0960, "000100000000"),
        fake_hidpp.Response("000200000000", 0x0960, "000200000000"),
    ),
    Setup(
        FeatureTest(settings_templates.ExtendedAdjustableDpi, {0: 0x64, 1: 0xE4}, {0: 0x164}, 2, offset=0x9),
        {
            common.NamedInt(0, "X"): common.NamedInts.list([0x064, 0x074, 0x084, 0x0A4, 0x0C4, 0x0E4, 0x0124, 0x0164, 0x01C4]),
            common.NamedInt(1, "Y"): common.NamedInts.list([0x064, 0x074, 0x084, 0x0A4, 0x0C4, 0x0E4, 0x0124, 0x0164]),
        },
        fake_hidpp.Response("000001", 0x0910, "00"),  # supports y direction, no lod
        fake_hidpp.Response("0000000064E0100084E02000C4E02000", 0x0920, "000000"),
        fake_hidpp.Response("000001E4E0400124E0400164E06001C4", 0x0920, "000001"),
        fake_hidpp.Response("00000000000000000000000000000000", 0x0920, "000002"),
        fake_hidpp.Response("0000000064E0100084E02000C4E02000", 0x0920, "000100"),
        fake_hidpp.Response("000001E4E0400124E040016400000000", 0x0920, "000101"),
        fake_hidpp.Response("000064007400E4007400", 0x0950),
        fake_hidpp.Response("00006400E400", 0x0960, "00006400E400"),
        fake_hidpp.Response("00016400E400", 0x0960, "00016400E400"),
    ),
    Setup(
        FeatureTest(settings_templates.ExtendedAdjustableDpi, {0: 0x64, 1: 0xE4, 2: 1}, {1: 0x164}, 2, offset=0x9),
        {
            common.NamedInt(0, "X"): common.NamedInts.list([0x064, 0x074, 0x084, 0x0A4, 0x0C4, 0x0E4, 0x0124, 0x0164, 0x01C4]),
            common.NamedInt(1, "Y"): common.NamedInts.list([0x064, 0x074, 0x084, 0x0A4, 0x0C4, 0x0E4, 0x0124, 0x0164]),
            common.NamedInt(2, "LOD"): common.NamedInts(LOW=0, MEDIUM=1, HIGH=2),
        },
        fake_hidpp.Response("000003", 0x0910, "00"),  # supports y direction and lod
        fake_hidpp.Response("0000000064E0100084E02000C4E02000", 0x0920, "000000"),
        fake_hidpp.Response("000001E4E0400124E0400164E06001C4", 0x0920, "000001"),
        fake_hidpp.Response("00000000000000000000000000000000", 0x0920, "000002"),
        fake_hidpp.Response("0000000064E0100084E02000C4E02000", 0x0920, "000100"),
        fake_hidpp.Response("000001E4E0400124E040016400000000", 0x0920, "000101"),
        fake_hidpp.Response("000064007400E4007401", 0x0950),
        fake_hidpp.Response("00006400E401", 0x0960, "00006400E401"),
        fake_hidpp.Response("000064016401", 0x0960, "000064016401"),
    ),
]


@pytest.mark.parametrize("test", key_tests)
def test_key_template(test, mocker):
    tst = test.test
    device = fake_hidpp.Device(responses=test.responses, feature=tst.sclass.feature, offset=tst.offset, version=tst.version)
    spy_request = mocker.spy(device, "request")

    print("FEATURE", tst.sclass.feature)
    setting = settings_templates.check_feature(device, tst.sclass)
    assert setting is not None
    if isinstance(setting, list):
        setting = setting[0]
    if isinstance(test.choices, list):
        assert setting._validator.min_value == test.choices[0]
        assert setting._validator.max_value == test.choices[1]
    elif test.choices is not None:
        assert setting.choices == test.choices

    value = setting.read(cached=False)
    assert value == tst.initial_value

    write_value = setting.write(value)
    assert write_value == tst.initial_value

    for key, val in tst.write_value.items():
        write_value = setting.write_key_value(key, val)
        assert write_value == val
        value[key] = val

    if tst.rewrite:
        write_value = setting.write(value)

    fake_hidpp.match_requests(tst.matched_calls, test.responses, spy_request.call_args_list)


@pytest.mark.parametrize(
    "responses, currentSpeed, newSpeed",
    [
        (fake_hidpp.responses_speedchange, 100, 200),
        (fake_hidpp.responses_speedchange, None, 250),
    ],
)
def test_SpeedChange_action(responses, currentSpeed, newSpeed, mocker):
    device = fake_hidpp.Device(responses=responses, feature=hidpp20_constants.SupportedFeature.POINTER_SPEED)
    spy_setting_callback = mocker.spy(device, "setting_callback")
    settings_templates.check_feature_settings(device, device.settings)  # need to set up all the settings
    device.persister = {"pointer_speed": currentSpeed, "_speed-change": newSpeed}
    speed_setting = next(filter(lambda s: s.name == "speed-change", device.settings), None)
    pointer_setting = next(filter(lambda s: s.name == "pointer_speed", device.settings), None)

    speed_setting.write(237)
    speed_setting._rw.press_action()

    if newSpeed is not None and speed_setting is not None:
        spy_setting_callback.assert_any_call(device, type(pointer_setting), [newSpeed])
    assert device.persister["_speed-change"] == currentSpeed
    assert device.persister["pointer_speed"] == newSpeed


@pytest.mark.parametrize("test", simple_tests + key_tests)
def test_check_feature_settings(test, mocker):
    tst = test.test
    device = fake_hidpp.Device(responses=test.responses, feature=tst.sclass.feature, offset=tst.offset, version=tst.version)

    already_known = []
    setting = settings_templates.check_feature_settings(device, already_known)

    assert setting is True
    assert already_known


@pytest.mark.parametrize(
    "test",
    [
        Setup(
            FeatureTest(settings_templates.K375sFnSwap, False, True, offset=0x06),
            fake_hidpp.Response("FF0001", 0x0600, "FF"),
            fake_hidpp.Response("FF0101", 0x0610, "FF01"),
        )
    ],
)
def test_check_feature_setting(test, mocker):
    tst = test.test
    device = fake_hidpp.Device(responses=test.responses, feature=tst.sclass.feature, offset=tst.offset, version=tst.version)

    setting = settings_templates.check_feature_setting(device, tst.sclass.name)

    assert setting



import struct
import sys

from typing import Union
from unittest import mock

import pytest

from logitech_receiver import base
from logitech_receiver import exceptions
from logitech_receiver.base import HIDPP_SHORT_MESSAGE_ID
from logitech_receiver.common import LOGITECH_VENDOR_ID
from logitech_receiver.common import BusID
from logitech_receiver.hidpp10_constants import ErrorCode as Hidpp10Error
from logitech_receiver.hidpp20_constants import ErrorCode as Hidpp20Error


@pytest.mark.parametrize(
    "usb_id, expected_name, expected_receiver_kind",
    [
        (0xC548, "Bolt Receiver", "bolt"),
        (0xC52B, "Unifying Receiver", "unifying"),
        (0xC531, "Nano Receiver", "nano"),
        (0xC53F, "Lightspeed Receiver", None),
        (0xC517, "EX100 Receiver 27 Mhz", "27Mhz"),
    ],
)
def test_product_information(usb_id, expected_name, expected_receiver_kind):
    res = base.product_information(usb_id)

    assert res["name"] == expected_name
    assert isinstance(res["vendor_id"], int)
    assert isinstance(res["product_id"], int)

    if expected_receiver_kind:
        assert res["receiver_kind"] == expected_receiver_kind


def test_filter_receivers_known():
    bus_id = 2
    product_id = 0xC548

    receiver_info = base.get_known_receiver_info(bus_id, LOGITECH_VENDOR_ID, product_id)

    assert receiver_info["name"] == "Bolt Receiver"
    assert receiver_info["receiver_kind"] == "bolt"


def test_filter_receivers_unknown():
    bus_id = 1
    product_id = 0xC500

    receiver_info = base.get_known_receiver_info(bus_id, LOGITECH_VENDOR_ID, product_id)

    assert receiver_info["bus_id"] == bus_id
    assert receiver_info["product_id"] == product_id


@pytest.mark.parametrize(
    "product_id, bus, hidpp_short, hidpp_long, expected",
    [
        (0xC548, BusID.USB, True, False, {"name": "Bolt Receiver", "usb_interface": 2}),
        (0xC07D, BusID.USB, True, False, {"usb_interface": 1}),
        (0xC07E, BusID.USB, False, True, {"usb_interface": 1}),
        (0xC07E, BusID.BLUETOOTH, False, True, {"bus_id": 5}),
        (0xA07E, BusID.USB, False, True, {"product_id": 0xA07E}),
        (0xA07C, BusID.USB, False, False, None),
        (0xC07F, BusID.USB, None, None, {"usb_interface": 2}),
        (0xC07F, BusID.BLUETOOTH, None, None, None),
        (0xB013, BusID.BLUETOOTH, None, None, {"product_id": 0xB013}),
    ],
)
def test_filter_products_of_interest(product_id, bus, hidpp_short, hidpp_long, expected):
    receiver_info = base.filter_products_of_interest(
        bus,
        LOGITECH_VENDOR_ID,
        product_id,
        hidpp_short=hidpp_short,
        hidpp_long=hidpp_long,
    )

    if expected is None:
        assert receiver_info == expected
    else:
        assert all([receiver_info[key] == expected_value for key, expected_value in expected.items()])
        assert receiver_info["vendor_id"] == LOGITECH_VENDOR_ID
        assert receiver_info["product_id"]


def test_match():
    record = {"vendor_id": LOGITECH_VENDOR_ID}

    res = base._match_device(record, 0, LOGITECH_VENDOR_ID, 0)

    assert res is True


@pytest.mark.parametrize(
    "report_id, sub_id, address, valid_notification",
    [
        (0x1, 0x72, 0x57, True),
        (0x1, 0x40, 0x63, True),
        (0x1, 0x40, 0x71, True),
        (0x1, 0x80, 0x71, False),
        (0x1, 0x00, 0x70, False),
        (0x20, 0x09, 0x71, False),
        (0x1, 0x37, 0x71, False),
    ],
)
def test_make_notification(report_id, sub_id, address, valid_notification):
    devnumber = 123
    data = bytes([sub_id, address, 0x02, 0x03, 0x04])

    result = base.make_notification(report_id, devnumber, data)

    if valid_notification:
        assert isinstance(result, base.HIDPPNotification)
        assert result.report_id == report_id
        assert result.devnumber == devnumber
        assert result.sub_id == sub_id
        assert result.address == address
        assert result.data == bytes([0x02, 0x03, 0x04])
    else:
        assert result is None


def test_get_next_sw_id():
    res1 = base._get_next_sw_id()
    res2 = base._get_next_sw_id()

    assert res1 == 2
    assert res2 == 3


@pytest.mark.parametrize(
    "prefix, error_code, return_error, raise_exception",
    [
        (b"\x8f", Hidpp10Error.INVALID_SUB_ID_COMMAND, False, False),
        (b"\x8f", Hidpp10Error.INVALID_SUB_ID_COMMAND, True, False),
        (b"\xff", Hidpp20Error.UNKNOWN, False, True),
    ],
)
def test_request_errors(
    prefix: bytes, error_code: Union[Hidpp10Error, Hidpp20Error], return_error: bool, raise_exception: bool
):
    handle = 0
    device_number = 66

    next_sw_id = 0x02
    reply_data_sw_id = struct.pack("!H", 0x0000 | next_sw_id)

    with mock.patch(
        "logitech_receiver.base._read",
        return_value=(HIDPP_SHORT_MESSAGE_ID, device_number, prefix + reply_data_sw_id + struct.pack("B", error_code)),
    ), mock.patch("logitech_receiver.base._read_input_buffer"), mock.patch(
        "logitech_receiver.base.write", return_value=None
    ), mock.patch("logitech_receiver.base._get_next_sw_id", return_value=next_sw_id):
        if raise_exception:
            with pytest.raises(exceptions.FeatureCallError) as context:
                base.request(handle, device_number, next_sw_id, return_error=return_error)
            assert context.value.number == device_number
            assert context.value.request == next_sw_id
            assert context.value.error == error_code
            assert context.value.params == b""

        else:
            result = base.request(handle, device_number, next_sw_id, return_error=return_error)
            assert result == (error_code if return_error else None)


@pytest.mark.skipif(sys.platform == "darwin", reason="Test only runs on Linux")
@pytest.mark.parametrize(
    "simulated_error, expected_result",
    [
        (Hidpp10Error.INVALID_SUB_ID_COMMAND, 1.0),
        (Hidpp10Error.RESOURCE_ERROR, None),
        (Hidpp10Error.CONNECTION_REQUEST_FAILED, None),
        (Hidpp10Error.UNKNOWN_DEVICE, exceptions.NoSuchDevice),
    ],
)
def test_ping_errors(simulated_error: Hidpp10Error, expected_result):
    handle = 1
    device_number = 1

    next_sw_id = 0x05
    reply_data_sw_id = struct.pack("!H", 0x0010 | next_sw_id)

    with mock.patch(
        "logitech_receiver.base._read",
        return_value=(HIDPP_SHORT_MESSAGE_ID, device_number, b"\x8f" + reply_data_sw_id + bytes([simulated_error])),
    ), mock.patch("logitech_receiver.base._get_next_sw_id", return_value=next_sw_id):
        if isinstance(expected_result, type) and issubclass(expected_result, Exception):
            with pytest.raises(expected_result) as context:
                base.ping(handle=handle, devnumber=device_number)
            assert context.value.number == device_number
            assert context.value.request == struct.unpack("!H", reply_data_sw_id)[0]

        else:
            result = base.ping(handle=handle, devnumber=device_number)
            assert result == expected_result



## Copyright (C) 2024  Solaar Contributors https://pwr-solaar.github.io/Solaar/
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""HID++ data and functions common to several logitech_receiver test files"""

from __future__ import annotations

import errno
import threading

from dataclasses import dataclass
from dataclasses import field
from struct import pack
from typing import Any
from typing import Optional

from logitech_receiver import device
from logitech_receiver import hidpp20
from solaar import configuration


def open_path(path: Optional[str]) -> int:
    if path is None:
        raise OSError(errno.EACCES, "Fake access error")
    return int(path, 16)  # can raise exception


def ping(responses, handle, devnumber, long_message=False):
    for r in responses:
        if handle == r.handle and devnumber == r.devnumber and r.id == 0x0010:
            return r.response


def request(
    responses,
    handle,
    devnumber,
    id,
    *params,
    no_reply=False,
    return_error=False,
    long_message=False,
    protocol=1.0,
):
    params = b"".join(pack("B", p) if isinstance(p, int) else p for p in params)
    print("REQUEST ", hex(handle), hex(devnumber), hex(id), params.hex())
    for r in responses:
        if handle == r.handle and devnumber == r.devnumber and r.id == id and bytes.fromhex(r.params) == params:
            print("RESPONSE", hex(r.handle), hex(r.devnumber), hex(r.id), r.params, r.response)
            return bytes.fromhex(r.response) if r.response is not None else None


@dataclass
class Response:
    response: str | float
    id: int
    params: str = ""
    handle: int = 0x11
    devnumber: int = 0xFF
    no_reply: bool = False


def replace_number(responses, number):  # change the devnumber for a list of responses
    return [Response(r.response, r.id, r.params, r.handle, number, r.no_reply) for r in responses]


def adjust_responses_index(index, responses):  # change index-4 responses to index
    return [Response(r.response, r.id - 0x400 + (index << 8), r.params, r.handle, r.devnumber, r.no_reply) for r in responses]


r_empty = [  # a HID++ device with no responses except for ping
    Response(1.0, 0x0010),  # ping
]

r_keyboard_1 = [  # a HID++ 1.0 keyboard
    Response(1.0, 0x0010),  # ping
    Response("001234", 0x81F1, "01"),  # firmware
    Response("003412", 0x81F1, "02"),  # firmware
    Response("002345", 0x81F1, "03"),  # firmware
    Response("003456", 0x81F1, "04"),  # firmware
    Response("050050", 0x8107),  # battery status
]

r_keyboard_2 = [  # a HID++ 2.0 keyboard
    Response(4.2, 0x0010),  # ping
    Response("010001", 0x0000, "0001"),  # feature set at 0x01
    Response("020003", 0x0000, "0020"),  # CONFIG_CHANGE at 0x02
    Response("030001", 0x0000, "0003"),  # device information at 0x03
    Response("040003", 0x0000, "0100"),  # unknown 0100 at 0x04
    Response("050003", 0x0000, "1B04"),  # reprogrammable keys V4 at 0x05
    Response("060003", 0x0000, "0007"),  # device friendly name at 0x06
    Response("070003", 0x0000, "0005"),  # device name at 0x07
    Response("080003", 0x0000, "1000"),  # battery status at 0x08
    Response("08", 0x0100),  # 8 features
    Response("00010001", 0x0110, "01"),  # feature set at 0x01
    Response("00200003", 0x0110, "02"),  # CONFIG_CHANGE at 0x02
    Response("00030001", 0x0110, "03"),  # device information at 0x03
    Response("01000003", 0x0110, "04"),  # unknown 0100 at 0x04
    Response("1B040003", 0x0110, "05"),  # reprogrammable keys V4 at 0x05
    Response("00070003", 0x0000, "06"),  # device friendly name at 0x06
    Response("00050003", 0x0000, "07"),  # device name at 0x07
    Response("10000001", 0x0110, "08"),  # battery status at 0x02
    Response("0212345678000D1234567890ABAA01", 0x0300),  # device information
    Response("04", 0x0500),  # reprogrammable keys V4
    Response("00110012AB010203CD00", 0x0510, "00"),  # reprogrammable keys V4
    Response("01110022AB010203CD00", 0x0510, "01"),  # reprogrammable keys V4
    Response("00010111AB010203CD00", 0x0510, "02"),  # reprogrammable keys V4
    Response("03110032AB010204CD00", 0x0510, "03"),  # reprogrammable keys V4
    Response("00030333AB010203CD00", 0x0510, "04"),  # reprogrammable keys V4
    Response("12", 0x0600),  # friendly namme
    Response("004142434445464748494A4B4C4D4E", 0x0610, "00"),
    Response("0E4F50515253000000000000000000", 0x0610, "0E"),
    Response("12", 0x0700),  # name and kind
    Response("4142434445464748494A4B4C4D4E4F", 0x0710, "00"),
    Response("505152530000000000000000000000", 0x0710, "0F"),
    Response("00", 0x0720),
    Response("12345678", 0x0800),  # battery status
]

r_mouse_1 = [  # a HID++ 1.0 mouse
    Response(1.0, 0x0010),  # ping
]

r_mouse_2 = [  # a HID++ 2.0 mouse with few responses except for ping
    Response(4.2, 0x0010),  # ping
]

r_mouse_3 = [  # a HID++ 2.0 mouse
    Response(4.5, 0x0010),  # ping
    Response("010001", 0x0000, "0001"),  # feature set at 0x01
    Response("020002", 0x0000, "8060"),  # report rate at 0x02
    Response("040001", 0x0000, "0003"),  # device information at 0x04
    Response("050002", 0x0000, "0005"),  # device type and name at 0x05
    Response("08", 0x0100),  # 8 features
    Response("00010001", 0x0110, "01"),  # feature set at 0x01
    Response("80600002", 0x0110, "02"),  # report rate at 0x02
    Response("00030001", 0x0110, "04"),  # device information at 0x04
    Response("00050002", 0x0110, "05"),  # device type and name at 0x05
    Response("09", 0x0210),  # report rate - current rate
    Response("03123456790008123456780000AA01", 0x0400),  # device information
    Response("0141424302030100", 0x0410, "00"),  # firmware 0
    Response("0241", 0x0410, "01"),  # firmware 1
    Response("05", 0x0410, "02"),  # firmware 2
    Response("12", 0x0500),  # name count - 18 characters
    Response("414241424142414241424142414241", 0x0510, "00"),  # name - first 15 characters
    Response("444544000000000000000000000000", 0x0510, "0F"),  # name - last 3 characters
]


responses_key = [  # responses for Reprogrammable Keys V4 at 0x05
    Response("08", 0x0500),  # Reprogrammable Keys V4 count
    Response("00500038010001010400000000000000", 0x0510, "00"),  # left button
    Response("00510039010001010400000000000000", 0x0510, "01"),  # right button
    Response("0052003A310003070500000000000000", 0x0510, "02"),  # middle button
    Response("0053003C710002030100000000000000", 0x0510, "03"),  # back button
    Response("0056003E710002030100000000000000", 0x0510, "04"),  # forward button
    Response("00C300A9310003070300000000000000", 0x0510, "05"),  # smart shift?
    Response("00C4009D310003070500000000000000", 0x0510, "06"),  # ?
    Response("00D700B4A00004000300000000000000", 0x0510, "07"),  # ?
    Response("00500000000000000000000000000000", 0x0520, "0050"),  # left button
    Response("00510000000000000000000000000000", 0x0520, "0051"),  # ...
    Response("00520100500000000000000000000000", 0x0520, "0052"),
    Response("00530500000000000000000000000000", 0x0520, "0053"),
    Response("00561100000000000000000000000000", 0x0520, "0056"),
    Response("00C30000000000000000000000000000", 0x0520, "00C3"),
    Response("00C40000500000000000000000000000", 0x0520, "00C4"),
    Response("00D70000510000000000000000000000", 0x0520, "00D7"),
    Response("0041", 0x0400),  # flags
    Response("0401", 0x0410),  # count
    Response("0050", 0x0420, "00FF"),  # left button
    Response("0051", 0x0420, "01FF"),  # right button
    Response("0052", 0x0420, "02FF"),  # middle button
    Response("0053", 0x0420, "03FF"),  # back button
    Response("0050000100500000", 0x0430, "0050FF"),  # left button current
    Response("0051000100500001", 0x0430, "0051FF"),  # right button current
    Response("0052000100500001", 0x0430, "0052FF"),  # middle button current
    Response("0053000100500001", 0x0430, "0053FF"),  # back button current
    Response("0050FF01005000", 0x0440, "0050FF01005000"),  # left button write
    Response("0051FF01005000", 0x0440, "0051FF01005000"),  # right button write
    Response("0051FF01005100", 0x0440, "0051FF01005100"),  # right button set write
]

responses_remap = [  # responses for Persistent Remappable Actions at 0x04 and reprogrammable keys at 0x05
    Response("0041", 0x0400),
    Response("03", 0x0410),
    Response("0301", 0x0410, "00"),
    Response("0050", 0x0420, "00FF"),
    Response("0050000200010001", 0x0430, "0050FF"),  # Left Button
    Response("0051", 0x0420, "01FF"),
    Response("0051000200010000", 0x0430, "0051FF"),  # Left Button
    Response("0052", 0x0420, "02FF"),
    Response("0052000100510000", 0x0430, "0052FF"),  # key DOWN
    Response("050002", 0x0000, "1B04"),  # REPROGRAMMABLE_KEYS_V4
] + responses_key

responses_gestures = [  # the commented-out messages are not used by either the setting or other testing
    Response("4203410141020400320480148C21A301", 0x0400, "0000"),  # items
    Response("A302A11EA30A4105822C852DAD2AAD2B", 0x0400, "0008"),
    Response("8F408F418F434204AF54912282558264", 0x0400, "0010"),
    Response("01000000000000000000000000000000", 0x0400, "0018"),
    Response("01000000000000000000000000000000", 0x0410, "000101"),  # enable
    #    Response("02000000000000000000000000000000", 0x0410, "000102"),
    #    Response("04000000000000000000000000000000", 0x0410, "000104"),
    #    Response("08000000000000000000000000000000", 0x0410, "000108"),
    Response("00000000000000000000000000000000", 0x0410, "000110"),
    #    Response("20000000000000000000000000000000", 0x0410, "000120"),
    #    Response("40000000000000000000000000000000", 0x0410, "000140"),
    #    Response("00000000000000000000000000000000", 0x0410, "000180"),
    #    Response("00000000000000000000000000000000", 0x0410, "010101"),
    #    Response("00000000000000000000000000000000", 0x0410, "010102"),
    #    Response("04000000000000000000000000000000", 0x0410, "010104"),
    #    Response("00000000000000000000000000000000", 0x0410, "010108"),
    Response("6F000000000000000000000000000000", 0x0410, "0001FF"),
    Response("04000000000000000000000000000000", 0x0410, "01010F"),
    Response("00000000000000000000000000000000", 0x0430, "000101"),  # divert
    #    Response("00000000000000000000000000000000", 0x0430, "000102"),
    #    Response("00000000000000000000000000000000", 0x0430, "000104"),
    #    Response("00000000000000000000000000000000", 0x0430, "000108"),
    Response("00000000000000000000000000000000", 0x0430, "000110"),
    #    Response("00000000000000000000000000000000", 0x0430, "000120"),
    #    Response("00000000000000000000000000000000", 0x0430, "000140"),
    #    Response("00000000000000000000000000000000", 0x0430, "000180"),
    #    Response("00000000000000000000000000000000", 0x0430, "010101"),
    #    Response("00000000000000000000000000000000", 0x0430, "010102"),
    Response("00000000000000000000000000000000", 0x0430, "0001FF"),
    Response("00000000000000000000000000000000", 0x0430, "010103"),
    Response("08000000000000000000000000000000", 0x0450, "01FF"),
    Response("08000000000000000000000000000000", 0x0450, "02FF"),
    Response("08000000000000000000000000000000", 0x0450, "03FF"),
    Response("00040000000000000000000000000000", 0x0450, "04FF"),
    Response("5C020000000000000000000000000000", 0x0450, "05FF"),
    Response("01000000000000000000000000000000", 0x0460, "00FF"),
    Response("01000000000000000000000000000000", 0x0470, "00FF"),
    Response("01", 0x0420, "00010101"),  # set index 1
    Response("00", 0x0420, "00010100"),  # unset index 1
    Response("01", 0x0420, "00011010"),  # set index 4
    Response("00", 0x0420, "00011000"),  # unset index 4
    Response("01", 0x0440, "00010101"),  # divert index 1
    Response("00", 0x0440, "00010100"),  # undivert index 1
    Response("000080FF", 0x0480, "000080FF"),  # write param 0
    Response("000180FF", 0x0480, "000180FF"),  # write param 0
]

zone_responses_1 = [  # responses for COLOR LED EFFECTS
    Response("00000102", 0x0710, "00FF00"),
    Response("0000000300040005", 0x0720, "000000"),
    Response("0001000B00080009", 0x0720, "000100"),
]
zone_responses_2 = [  # responses for RGB EFFECTS
    Response("0000000102", 0x0700, "00FF00"),
    Response("0000000300040005", 0x0700, "000000"),
    Response("0001000200080009", 0x0700, "000100"),
]
effects_responses_1 = [Response("0100000001", 0x0700)] + zone_responses_1
effects_responses_2 = [Response("FFFF0100000001", 0x0700, "FFFF00")] + zone_responses_2

responses_profiles = [  # OnboardProfile in RAM
    Response("0104010101020100FE0200", 0x0900),
    Response("000101FF", 0x0950, "00000000"),
    Response("FFFFFFFF", 0x0950, "00000004"),
    Response("01010290018003000700140028FFFFFF", 0x0950, "00010000"),
    Response("FFFF0000000000000000000000000000", 0x0950, "00010010"),
    Response("8000FFFF900aFF00800204548000FFFF", 0x0950, "00010020"),
    Response("900aFF00800204548000FFFF900aFF00", 0x0950, "00010030"),
    Response("800204548000FFFF900aFF0080020454", 0x0950, "00010040"),
    Response("8000FFFF900aFF00800204548000FFFF", 0x0950, "00010050"),
    Response("FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF", 0x0950, "00010060"),
    Response("FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF", 0x0950, "00010070"),
    Response("FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF", 0x0950, "00010080"),
    Response("FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF", 0x0950, "00010090"),
    Response("54004500370000000000000000000000", 0x0950, "000100A0"),
    Response("00000000000000000000000000000000", 0x0950, "000100B0"),
    Response("00000000000000000000000000000000", 0x0950, "000100C0"),
    Response("0A01020300500407000000FFFFFFFFFF", 0x0950, "000100D0"),
    Response("FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF", 0x0950, "000100E0"),
    Response("FFFFFFFFFFFFFFFFFFFFFFFFFF7C81AB", 0x0950, "000100EE"),
]
responses_profiles_rom = [  # OnboardProfile in ROM
    Response("0104010101020100FE0200", 0x0900),
    Response("00000000", 0x0950, "00000000"),
    Response("010101FF", 0x0950, "01000000"),
    Response("FFFFFFFF", 0x0950, "01000004"),
    Response("01010290018003000700140028FFFFFF", 0x0950, "01010000"),
    Response("FFFF0000000000000000000000000000", 0x0950, "01010010"),
    Response("8000FFFF900aFF00800204548000FFFF", 0x0950, "01010020"),
    Response("900aFF00800204548000FFFF900aFF00", 0x0950, "01010030"),
    Response("800204548000FFFF900aFF0080020454", 0x0950, "01010040"),
    Response("8000FFFF900aFF00800204548000FFFF", 0x0950, "01010050"),
    Response("FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF", 0x0950, "01010060"),
    Response("FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF", 0x0950, "01010070"),
    Response("FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF", 0x0950, "01010080"),
    Response("FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF", 0x0950, "01010090"),
    Response("54004500370000000000000000000000", 0x0950, "010100A0"),
    Response("00000000000000000000000000000000", 0x0950, "010100B0"),
    Response("00000000000000000000000000000000", 0x0950, "010100C0"),
    Response("0A01020300500407000000FFFFFFFFFF", 0x0950, "010100D0"),
    Response("FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF", 0x0950, "010100E0"),
    Response("FFFFFFFFFFFFFFFFFFFFFFFFFF7C81AB", 0x0950, "010100EE"),
]
responses_profiles_rom_2 = [  # OnboardProfile in ROM
    Response("0104010101020100FE0200", 0x0900),
    Response("FFFFFFFF", 0x0950, "00000000"),
    Response("010101FF", 0x0950, "01000000"),
    Response("FFFFFFFF", 0x0950, "01000004"),
    Response("01010290018003000700140028FFFFFF", 0x0950, "01010000"),
    Response("FFFF0000000000000000000000000000", 0x0950, "01010010"),
    Response("8000FFFF900aFF00800204548000FFFF", 0x0950, "01010020"),
    Response("900aFF00800204548000FFFF900aFF00", 0x0950, "01010030"),
    Response("800204548000FFFF900aFF0080020454", 0x0950, "01010040"),
    Response("8000FFFF900aFF00800204548000FFFF", 0x0950, "01010050"),
    Response("FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF", 0x0950, "01010060"),
    Response("FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF", 0x0950, "01010070"),
    Response("FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF", 0x0950, "01010080"),
    Response("FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF", 0x0950, "01010090"),
    Response("54004500370000000000000000000000", 0x0950, "010100A0"),
    Response("00000000000000000000000000000000", 0x0950, "010100B0"),
    Response("00000000000000000000000000000000", 0x0950, "010100C0"),
    Response("0A01020300500407000000FFFFFFFFFF", 0x0950, "010100D0"),
    Response("FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF", 0x0950, "010100E0"),
    Response("FFFFFFFFFFFFFFFFFFFFFFFFFF7C81AB", 0x0950, "010100EE"),
]

complex_responses_1 = [  # COLOR_LED_EFFECTS
    Response(4.2, 0x0010),  # ping
    Response("010001", 0x0000, "0001"),  # FEATURE SET at x01
    Response("020001", 0x0000, "0020"),  # CONFIG_CHANGE at x02
    Response("0A", 0x0100),  # 10 features
    Response("070001", 0x0000, "8070"),  # COLOR_LED_EFFECTS at 0x07
    *effects_responses_1,
]

complex_responses_2 = [  # RGB_EFFECTS + reprogrammable keys + persistent actions
    Response(4.2, 0x0010),  # ping
    Response("010001", 0x0000, "0001"),  # FEATURE SET at x01
    Response("020001", 0x0000, "0020"),  # CONFIG_CHANGE at x02
    Response("0A", 0x0100),  # 10 features
    Response("070001", 0x0000, "8071"),  # RGB_EFFECTS at 0x07
    *effects_responses_2,
    Response("040001", 0x0000, "1C00"),  # Persistent Remappable Actions at 0x04
    *responses_remap,
    Response("080001", 0x0000, "6501"),  # Gestures at 0x08
    *adjust_responses_index(8, responses_gestures),
    Response("060003", 0x0000, "1982"),  # Backlight 2 at 0x06
    Response("010118000001020003000400", 0x0600),
    Response("090003", 0x0000, "8100"),  # Onboard Profiles at 0x09
    *responses_profiles,
]

responses_speedchange = [
    Response("0100", 0x0400),
    Response("010001", 0x0000, "0001"),  # FEATURE SET at x01
    Response("0A", 0x0100),  # 10 features
    Response("0120", 0x0410, "0120"),
    Response("050001", 0x0000, "1B04"),  # REPROG_CONTROLS_V4
    Response("01", 0x0500),
    Response("00ED009D310003070500000000000000", 0x0510, "00"),  # DPI Change
    Response("00ED0000000000000000000000000000", 0x0520, "00ED"),  # DPI Change current
    Response("060000", 0x0000, "2205"),  # POINTER_SPEED
]


# A fake device that uses provided data (responses) to respond to HID++ commands.
# Some methods from the real device are used to set up data structures needed for settings
@dataclass
class Device:
    name: str = "TESTD"
    online: bool = True
    protocol: float = 2.0
    responses: Any = field(default_factory=list)
    codename: str = "TESTC"
    feature: Optional[int] = None
    offset: Optional[int] = 4
    version: Optional[int] = 0
    wpid: Optional[str] = "0000"
    setting_callback: Any = None
    sliding = profiles = _backlight = _keys = _remap_keys = _led_effects = _gestures = None
    _gestures_lock = threading.Lock()
    number = "d1"
    present = True

    read_register = device.Device.read_register
    write_register = device.Device.write_register
    backlight = device.Device.backlight
    keys = device.Device.keys
    remap_keys = device.Device.remap_keys
    led_effects = device.Device.led_effects
    gestures = device.Device.gestures
    __hash__ = device.Device.__hash__
    feature_request = device.Device.feature_request

    def __post_init__(self):
        self._name = self.name
        self._protocol = self.protocol
        self.persister = configuration._DeviceEntry()
        self.features = hidpp20.FeaturesArray(self)
        self.settings = []
        self.receiver = []
        if self.feature is not None:
            self.features = hidpp20.FeaturesArray(self)
            self.responses = [
                Response("010001", 0x0000, "0001"),
                Response("20", 0x0100),
            ] + self.responses
            self.responses.append(
                Response(
                    f"{int(self.offset):0>2X}00{int(self.version):0>2X}",
                    0x0000,
                    f"{int(self.feature):0>4X}",
                )
            )
        if self.setting_callback is None:
            self.setting_callback = lambda x, y, z: None
        self.add_notification_handler = lambda x, y: None

    def request(self, id, *params, no_reply=False, long_message=False, protocol=2.0):
        params = b"".join(pack("B", p) if isinstance(p, int) else p for p in params)
        print("REQUEST ", self._name, hex(id), params.hex().upper())
        for r in self.responses:
            if id == r.id and params == bytes.fromhex(r.params):
                print("RESPONSE", self._name, hex(r.id), r.params, r.response)
                return bytes.fromhex(r.response) if isinstance(r.response, str) else r.response
        print("RESPONSE", self._name, None)

    def ping(self, handle=None, devnumber=None, long_message=False):
        print("PING", self._protocol)
        return self._protocol

    def handle_notification(self, handle):
        pass

    def changed(self, *args, **kwargs):
        pass

    def set_battery_info(self, *args, **kwargs):
        pass

    def status_string(self):
        pass


def match_requests(number, responses, call_args_list):
    for i in range(0 - number, 0):
        param = b"".join(pack("B", p) if isinstance(p, int) else p for p in call_args_list[i][0][1:]).hex().upper()
        print("MATCH", i, hex(call_args_list[i][0][0]), param, hex(responses[i].id), responses[i].params)
        assert call_args_list[i][0][0] == responses[i].id
        assert param == responses[i].params



import subprocess
import textwrap

from glob import glob
from os.path import dirname
from pathlib import Path

from setuptools import find_packages
from setuptools import setup

NAME = "Solaar"
version = Path("lib/solaar/version").read_text().strip()

try:  # get commit from git describe
    commit = subprocess.check_output(["git", "describe", "--always"], stderr=subprocess.DEVNULL).strip().decode()
    Path("lib/solaar/commit").write_text(f"{commit}\n")
except Exception:  # get commit from Ubuntu dpkg-parsechangelog
    try:
        commit = (
            subprocess.check_output(["dpkg-parsechangelog", "--show-field", "Version"], stderr=subprocess.DEVNULL)
            .strip()
            .decode()
        )
        commit = commit.split("~")
        Path("lib/solaar/commit").write_text(f"{commit[0]}\n")
    except Exception as e:
        print("Exception using dpkg-parsechangelog", e)


def _data_files():
    yield "share/icons/hicolor/scalable/apps", glob("share/solaar/icons/solaar*.svg")
    yield "share/icons/hicolor/32x32/apps", glob("share/solaar/icons/solaar-light_*.png")

    for mo in glob("share/locale/*/LC_MESSAGES/solaar.mo"):
        yield dirname(mo), [mo]

    yield "share/applications", ["share/applications/solaar.desktop"]
    yield "lib/udev/rules.d", ["rules.d/42-logitech-unify-permissions.rules"]
    yield "share/metainfo", ["share/solaar/io.github.pwr_solaar.solaar.metainfo.xml"]


setup(
    name=NAME.lower(),
    version=version,
    description="Linux device manager for Logitech receivers, keyboards, mice, and tablets.",
    long_description=textwrap.dedent(
        """
        Solaar is a Linux device manager for many Logitech peripherals that connect through
        Unifying and other receivers or via USB or Bluetooth.
        Solaar is able to pair/unpair devices with receivers and show and modify some of the
        modifiable features of devices.
        For instructions on installing Solaar see https://pwr-solaar.github.io/Solaar/installation"""
    ),
    author="Daniel Pavel",
    license="GPLv2",
    url="http://pwr-solaar.github.io/Solaar/",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: X11 Applications :: GTK",
        "Environment :: Console",
        "Intended Audience :: End Users/Desktop",
        "License :: DFSG approved",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Natural Language :: English",
        "Programming Language :: Python :: 3 :: Only",
        "Operating System :: POSIX :: Linux",
        "Topic :: Utilities",
    ],
    platforms=["linux"],
    python_requires=">=3.8",
    install_requires=[
        'evdev (>= 1.1.2) ; platform_system=="Linux"',
        "pyudev (>= 0.13)",
        "PyYAML (>= 3.12)",
        "python-xlib (>= 0.27)",
        "psutil (>= 5.4.3)",
        'dbus-python ; platform_system=="Linux"',
        "PyGObject",
        "typing_extensions",
    ],
    extras_require={
        "report-descriptor": ["hid-parser"],
        "desktop-notifications": ["Notify (>= 0.7)"],
        "git-commit": ["python-git-info"],
        "test": ["pytest", "pytest-mock", "pytest-cov"],
        "dev": ["ruff"],
    },
    package_dir={"": "lib"},
    packages=find_packages(where="lib"),
    data_files=list(_data_files()),
    include_package_data=True,
    scripts=glob("bin/*"),
)



from dataclasses import dataclass
from functools import partial
from unittest import mock

import pytest

from logitech_receiver import base
from logitech_receiver import common
from logitech_receiver import exceptions
from logitech_receiver import receiver

from . import fake_hidpp


@pytest.fixture
def nano_recv():
    device_info = DeviceInfo("12", product_id=0xC534)
    mock_low_level = LowLevelInterfaceFake(responses_lacking)
    yield receiver.create_receiver(mock_low_level, device_info, lambda x: x)


class LowLevelInterfaceFake:
    def __init__(self, responses=None):
        self.responses = responses

    def open_path(self, path):
        return fake_hidpp.open_path(path)

    def product_information(self, usb_id: int) -> dict:
        return base.product_information(usb_id)

    def find_paired_node(self, receiver_path: str, index: int, timeout: int):
        return None

    def request(self, response, *args, **kwargs):
        func = partial(fake_hidpp.request, self.responses)
        return func(response, *args, **kwargs)

    def ping(self, response, *args, **kwargs):
        func = partial(fake_hidpp.ping, self.responses)
        return func(response, *args, **kwargs)

    def close(self, *args, **kwargs):
        pass


@pytest.mark.parametrize(
    "index, expected_kind",
    [
        (0, None),
        (1, 2),  # mouse
        (2, 2),  # mouse
        (3, 1),  # keyboard
        (4, 3),  # numpad
        (5, None),
    ],
)
def test_get_kind_from_index(index, expected_kind):
    mock_receiver = mock.Mock()

    if expected_kind:
        assert receiver._get_kind_from_index(mock_receiver, index) == expected_kind
    else:
        with pytest.raises(exceptions.NoSuchDevice):
            receiver._get_kind_from_index(mock_receiver, index)


@dataclass
class DeviceInfo:
    path: str
    vendor_id: int = 1133
    product_id: int = 0xC52B


responses_unifying = [
    fake_hidpp.Response("000000", 0x8003, "FF"),
    fake_hidpp.Response("000300", 0x8102),
    fake_hidpp.Response("0316CC9CB40506220000000000000000", 0x83B5, "03"),
    fake_hidpp.Response("20200840820402020700000000000000", 0x83B5, "20"),
    fake_hidpp.Response("21211420110400010D1A000000000000", 0x83B5, "21"),
    fake_hidpp.Response("22220840660402010700000000020000", 0x83B5, "22"),
    fake_hidpp.Response("30198E3EB80600000001000000000000", 0x83B5, "30"),
    fake_hidpp.Response("31811119511A40000002000000000000", 0x83B5, "31"),
    fake_hidpp.Response("32112C46EA1E40000003000000000000", 0x83B5, "32"),
    fake_hidpp.Response("400B4D58204D61737465722033000000", 0x83B5, "40"),
    fake_hidpp.Response("41044B35323020202020202020202020", 0x83B5, "41"),
    fake_hidpp.Response("42054372616674000000000000000000", 0x83B5, "42"),
    fake_hidpp.Response("012411", 0x81F1, "01"),
    fake_hidpp.Response("020036", 0x81F1, "02"),
    fake_hidpp.Response("03AAAC", 0x81F1, "03"),
    fake_hidpp.Response("040209", 0x81F1, "04"),
]
responses_c534 = [
    fake_hidpp.Response("000000", 0x8003, "FF", handle=0x12),
    fake_hidpp.Response("000209", 0x8102, handle=0x12),
    fake_hidpp.Response("0316CC9CB40502220000000000000000", 0x83B5, "03", handle=0x12),
    fake_hidpp.Response("00000445AB", 0x83B5, "04", handle=0x12),
]
responses_unusual = [
    fake_hidpp.Response("000000", 0x8003, "FF", handle=0x13),
    fake_hidpp.Response("000300", 0x8102, handle=0x13),
    fake_hidpp.Response("00000445AB", 0x83B5, "04", handle=0x13),
    fake_hidpp.Response("0326CC9CB40508220000000000000000", 0x83B5, "03", handle=0x13),
]
responses_lacking = [
    fake_hidpp.Response("000000", 0x8003, "FF", handle=0x14),
    fake_hidpp.Response("000300", 0x8102, handle=0x14),
]

mouse_info = {
    "kind": common.NamedInt(2, "mouse"),
    "polling": "8ms",
    "power_switch": common.NamedInt(1, "base"),
    "serial": "198E3EB8",
    "wpid": "4082",
}
c534_info = {"kind": common.NamedInt(0, "unknown"), "polling": "", "power_switch": "(unknown)", "serial": None, "wpid": "45AB"}


@pytest.mark.parametrize(
    "device_info, responses, handle, serial, max_devices, ",
    [
        (DeviceInfo(path=None), [], False, None, None),
        (DeviceInfo(path=11), [], None, None, None),
        (DeviceInfo(path="11"), responses_unifying, 0x11, "16CC9CB4", 6),
        (DeviceInfo(path="12", product_id=0xC534), responses_c534, 0x12, "16CC9CB4", 2),
        (DeviceInfo(path="12", product_id=0xC539), responses_c534, 0x12, "16CC9CB4", 2),
        (DeviceInfo(path="13"), responses_unusual, 0x13, "26CC9CB4", 1),
        (DeviceInfo(path="14"), responses_lacking, 0x14, None, 1),
    ],
)
def test_receiver_factory_create_receiver(device_info, responses, handle, serial, max_devices):
    mock_low_level = LowLevelInterfaceFake(responses)

    if handle is False:
        with pytest.raises(Exception):  # noqa: B017
            receiver.create_receiver(mock_low_level, device_info, lambda x: x)
    elif handle is None:
        r = receiver.create_receiver(mock_low_level, device_info, lambda x: x)
        assert r is None
    else:
        r = receiver.create_receiver(mock_low_level, device_info, lambda x: x)
        assert r.handle == handle
        assert r.serial == serial
        assert r.max_devices == max_devices


@pytest.mark.parametrize(
    "device_info, responses, firmware, codename, remaining_pairings, pairing_info, count",
    [
        (DeviceInfo("11"), responses_unifying, 3, "K520", -1, mouse_info, 3),
        (DeviceInfo("12", product_id=0xC534), responses_c534, None, None, 4, c534_info, 2),
        (DeviceInfo("13", product_id=0xCCCC), responses_unusual, None, None, -1, c534_info, 3),
    ],
)
def test_receiver_factory_props(device_info, responses, firmware, codename, remaining_pairings, pairing_info, count):
    mock_low_level = LowLevelInterfaceFake(responses)

    r = receiver.create_receiver(mock_low_level, device_info, lambda x: x)

    assert len(r.firmware) == firmware if firmware is not None else firmware is None
    assert r.device_codename(2) == codename
    assert r.remaining_pairings() == remaining_pairings
    assert r.device_pairing_information(1) == pairing_info
    assert r.count() == count


@pytest.mark.parametrize(
    "device_info, responses, status_str, strng",
    [
        (DeviceInfo("11"), responses_unifying, "No paired devices.", "<UnifyingReceiver(11,17)>"),
        (DeviceInfo("12", product_id=0xC534), responses_c534, "No paired devices.", "<NanoReceiver(12,18)>"),
        (DeviceInfo("13", product_id=0xCCCC), responses_unusual, "No paired devices.", "<Receiver(13,19)>"),
    ],
)
def test_receiver_factory_string(device_info, responses, status_str, strng):
    mock_low_level = LowLevelInterfaceFake(responses)

    r = receiver.create_receiver(mock_low_level, device_info, lambda x: x)

    assert r.status_string() == status_str
    assert str(r) == strng


@pytest.mark.parametrize(
    "device_info, responses",
    [
        (DeviceInfo("14"), responses_lacking),
        (DeviceInfo("14", product_id="C534"), responses_lacking),
    ],
)
def test_receiver_factory_no_device(device_info, responses):
    mock_low_level = LowLevelInterfaceFake(responses)

    r = receiver.create_receiver(mock_low_level, device_info, lambda x: x)

    with pytest.raises(exceptions.NoSuchDevice):
        r.device_pairing_information(1)


@pytest.mark.parametrize(
    "address, data, expected_online, expected_encrypted",
    [
        (0x03, b"\x01\x02\x03", True, False),
        (0x10, b"\x61\x02\x03", False, True),
    ],
)
def test_notification_information_nano_receiver(nano_recv, address, data, expected_online, expected_encrypted):
    _number = 0
    notification = base.HIDPPNotification(
        report_id=0x01,
        devnumber=0x52C,
        sub_id=0,
        address=address,
        data=data,
    )
    online, encrypted, wpid, kind = nano_recv.notification_information(_number, notification)

    assert online == expected_online
    assert encrypted == expected_encrypted
    assert wpid == "0302"
    assert kind == "keyboard"


def test_extract_serial_number():
    response = b'\x03\x16\xcc\x9c\xb4\x05\x06"\x00\x00\x00\x00\x00\x00\x00\x00'

    serial_number = receiver.extract_serial(response[1:5])

    assert serial_number == "16CC9CB4"


def test_extract_max_devices():
    response = b'\x03\x16\xcc\x9c\xb4\x05\x06"\x00\x00\x00\x00\x00\x00\x00\x00'

    max_devices = receiver.extract_max_devices(response)

    assert max_devices == 6


@pytest.mark.parametrize(
    "response, expected_remaining_pairings",
    [
        (b"\x00\x03\x00", -1),
        (b"\x00\x02\t", 4),
    ],
)
def test_extract_remaining_pairings(response, expected_remaining_pairings):
    remaining_pairings = receiver.extract_remaining_pairings(response)

    assert remaining_pairings == expected_remaining_pairings


def test_extract_codename():
    response = b"A\x04K520"

    codename = receiver.extract_codename(response)

    assert codename == "K520"


@pytest.mark.parametrize(
    "power_switch_byte, expected_location",
    [
        (b"\x01", "base"),
        (b"\x09", "top_edge"),
        (b"\x0c", "bottom_edge"),
        (b"\x00", "unknown"),
        (b"\x0f", "unknown"),
    ],
)
def test_extract_power_switch_location(power_switch_byte, expected_location):
    response = b"\x19\x8e>\xb8\x06\x00\x00\x00\x00" + power_switch_byte + b"\x00\x00\x00\x00\x00"

    ps_location = receiver.extract_power_switch_location(response)

    assert ps_location == expected_location


def test_extract_connection_count():
    response = b"\x00\x03\x00"

    connection_count = receiver.extract_connection_count(response)

    assert connection_count == 3


def test_extract_wpid():
    response = b"@\x82"

    res = receiver.extract_wpid(response)

    assert res == "4082"


def test_extract_polling_rate():
    response = b"\x08@\x82\x04\x02\x02\x07\x00\x00\x00\x00\x00\x00\x00"

    polling_rate = receiver.extract_polling_rate(response)

    assert polling_rate == 130


@pytest.mark.parametrize(
    "data, expected_device_kind",
    [
        (0x00, "unknown"),
        (0x03, "numpad"),
    ],
)
def test_extract_device_kind(data, expected_device_kind):
    device_kind = receiver.extract_device_kind(data)

    assert str(device_kind) == expected_device_kind



from unittest import mock

from logitech_receiver.hidpp10_constants import ErrorCode
from logitech_receiver.hidpp10_constants import Registers
from solaar.cli.probe import run


# Mock receiver class
class MockReceiver:
    handle = 1
    isDevice = False

    def read_register(self, register, *args):
        return 0 if register == Registers.RECEIVER_INFO else b"\x01\x03"


def test_run_register_errors():
    mock_args = mock.Mock()
    mock_args.receiver = False

    mock_receiver = MockReceiver()

    # Define expected addresses to be called in order
    expected_addresses = []

    for reg in range(0, 0xFF):
        expected_addresses.append((0x8100 | reg, 0))  # First short call, returns invalid_value (continue)
        expected_addresses.append((0x8100 | reg, 1))  # Second short call, returns invalid_address (stop here)

        expected_addresses.append((0x8100 | (0x200 + reg), 0))  # First long call, returns invalid_value (continue)
        expected_addresses.append((0x8100 | (0x200 + reg), 1))  # Second long call, returns invalid_address (stop here)

    # To record the actual addresses called
    called_addresses = []

    def mock_base_request(handle, devnumber, reg, sub, return_error=False):
        called_addresses.append((reg, sub))
        if sub == 0:
            return ErrorCode.INVALID_VALUE
        elif sub == 1:
            return ErrorCode.INVALID_ADDRESS
        return b"\x01\x02"

    with mock.patch("logitech_receiver.base.request", side_effect=mock_base_request), mock.patch(
        "solaar.cli.probe._print_receiver", return_value=None
    ):
        # Call the run function with mocked receivers and args (passing real find_receiver function)
        run([mock_receiver], mock_args, None, None)

        # Evaluate that the addresses called match the expected addresses
        assert (
            called_addresses == expected_addresses
        ), f"Called addresses {called_addresses} do not match expected {expected_addresses}"
